import subprocess


def Run_Audacity(audio_path):
    try:
        # Run the dpkg command to check if the package is installed
        subprocess.run(['audacity', str(audio_path)],
                       stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except subprocess.CalledProcessError:
        return False



"""
Helper for manim-recorder
"""

import json
import os
import textwrap
import datetime
import platform  # Added import for platform
import logging
from manim import logger, Mobject


def Check_os():
    """
    Check the operating system and return its name.

    Returns:
        str: The name of the operating system (e.g., "Linux", "macOS", "Windows", "Termux (Android)", or "Unknown OS").
    """
    os_name = platform.system()
    match os_name:
        case "Linux":
            # Check for Termux by looking for the presence of the 'termux' directory
            if os.path.exists("/data/data/com.termux"):
                return "Termux (Android)"
            else:
                return "Linux"
        case "Darwin":
            return "macOS"
        case "Windows":
            return "Windows"
        case _:
            return "Unknown OS"


def get_audio_basename() -> str:
    """
    Generates a unique audio file name based on the current date and time.

    Returns:
        str: The generated audio file name.
    """
    now = datetime.datetime.now()
    return "Voice_{}".format(now.strftime("%d%m%Y_%H%M%S"))


def mobject_to_text(mobject: Mobject) -> str():
    # Tex() or MathTex()
    if hasattr(mobject, "get_tex_string"):
        mobject = mobject.get_tex_string()
    # Text
    elif hasattr(mobject, "original_text"):
        mobject = mobject.original_text
    # Paragraph
    elif hasattr(mobject, "lines_text"):
        mobject = mobject.lines_text.original_text
    else:
        mobject = str(mobject)
    return mobject


def text_and_mobject(text: str, mobject: Mobject, **kwargs) -> tuple():
    match [text, mobject]:
        case [None, Mobject()]:
            text = mobject_to_text(mobject)
            mobject = (
                mobject.get_file_path() if hasattr(mobject, "get_file_path") else None
            )
        case [str(), Mobject()]:
            try:
                mobject = (
                    mobject.get_file_path()
                    if hasattr(mobject, "get_file_path")
                    else None
                )
                
            except AttributeError as e:
                logging.error("mobject : {}".format(e))
                mobject = None
        case [None, str()]:
            text = mobject
            mobject = mobject if os.path.exists(mobject) else None
        case [Mobject(), None] | [Mobject(), str()] | [str(), None]:
            return text_and_mobject(mobject, text)
        case _:
            return None
    
    return text, mobject


def msg_box(msg, indent=1, width=None, title=None):
    """
    Create a message box with optional title and indentation.

    Args:
        msg (str): The message to display in the box.
        indent (int): The number of spaces to indent the message (default is 1).
        width (int, optional): The width of the message box. If None, it will be determined based on the message length.
        title (str, optional): The title of the message box.

    Returns:
        str: A formatted string representing the message box.
    """
    # Wrap lines that are longer than 80 characters
    if width is None and len(msg) > 80:
        width = 80
        lines = []
        for line in msg.splitlines():
            if len(line) > width:
                line = line[:width] + " " + line[width:]
            lines.extend(textwrap.wrap(line, width))
        msg = "\n".join(lines)

    lines = msg.split("\n")
    space = " " * indent
    if not width:
        width = max(map(len, lines))
    box = f'╔{"═" * (width + indent * 2)}╗\n'  # upper_border
    if title:
        box += f"║{space}{title:<{width}}{space}║\n"  # title
        box += f'║{space}{"-" * len(title):<{width}}{space}║\n'  # underscore
    box += "".join([f"║{space}{line:<{width}}{space}║\n" for line in lines])
    box += f'╚{"═" * (width + indent * 2)}╝'  # lower_border
    return box


def append_to_json_file(json_file: str, data: dict, voice_id: int = -1, **kwargs):
    """
    Append data to a JSON file. If the file does not exist, it will be created.

    Args:
        json_file (str): The path to the JSON file.
        data (dict): The data to append to the JSON file.
        voice_id (int, optional): The index of the voice data to update (default is -1, which appends).
        **kwargs: Additional keyword arguments (not used in this function).

    Raises:
        ValueError: If the JSON file does not contain a list.
    """
    # This cache.json file is not exist and Create cache.json file and append
    if not os.path.exists(json_file):
        with open(json_file, "w") as f:
            json.dump([data], f, indent=2)
        return

    # This cache.json file is exist and load
    with open(json_file, "r") as f:
        json_data = json.load(f)

    # Check cache.json file is list
    if not isinstance(json_data, list):
        raise ValueError("JSON file should be a list")

    if voice_id > -1 and 0 <= voice_id < len(json_data):
        cache_json_data = json_data[voice_id]
        if isinstance(cache_json_data, dict):
            if data.get("input_data") == cache_json_data.get("input_data"):
                return
            elif isinstance(cache_json_data.get("original_audio"), str):
                final_audio = cache_json_data.get("original_audio")
                final_audio = "{}/{}".format(os.path.dirname(json_file), final_audio)
                if os.path.exists(final_audio):
                    os.remove(final_audio)

        json_data[voice_id] = data
        return
    else:
        json_data.append(data)

    with open(json_file, "w") as f:
        json.dump(json_data, f, indent=2, ensure_ascii=False)
    return


def create_dotenv_file(required_variable_names: list, dotenv=".env"):
    """
    Create a .env file with the required environment variables.

    Args:
        required_variable_names (list): A list of variable names to include in the .env file.
        dotenv (str, optional): The name of the .env file to create (default is ".env").

    Returns:
        bool: True if the .env file was created, False if it was skipped.
    """
    if os.path.exists(dotenv):
        logger.info(
            "File {} already exists. Would you like to overwrite it? [Y/n]".format(
                dotenv
            )
        )
        answer = input()
        if answer.lower() == "n":
            logger.info("Skipping .env file creation...")
            return False

    logger.info("Creating .env file...")
    with open(dotenv, "w") as f:
        for var_name in required_variable_names:
            logger.info(f"Enter value for {var_name}:")
            value = input()
            f.write(f"{var_name}={value}\n")

    return True



"""Manim Recorder plugin for manim"""

from manim_recorder.voiceover_scene import RecorderScene
from manim_recorder.tracker import SoundTracker

# import pkg_resources

# __version__: str = pkg_resources.get_distribution(__name__).version

# __version__ = "0.0.1"



"""
"""

import re
import os
from typing import override
from pathlib import Path
from math import ceil
from contextlib import contextmanager
from typing import Optional, Generator
from manim import Scene, config, Mobject, logger
from manim.utils.file_ops import open_media_file
from manim_recorder.recorder.base import AudioService
from manim_recorder.tracker import SoundTracker
from manim_recorder.multimedia import chunks
from manim_recorder.helper import text_and_mobject


class RecorderScene(Scene):
    """A scene class that can be used to add sound to a scene."""

    audio_service: AudioService
    current_tracker: Optional[SoundTracker]
    create_subcaption: bool
    create_script: bool
    voice_id: int = -1

    def set_audio_service(
        self,
        audio_service: AudioService,
        create_subcaption: bool | None = True,
        cache_dir_delete: bool = False,
    ) -> None:
        """Sets the Audio service to be used for the sound. This method
        should be called before adding any sound to the scene.

        Args:
            audio_service (AudioService): The audio service to be used.
            create_subcaption (bool, optional): Whether to create subcaptions for the scene. Defaults to True.
        """
        if hasattr(audio_service, "default_cache_dir"):
            if audio_service.default_cache_dir:
                audio_service.recording_cache_dir(
                    os.path.join(str(config.media_dir), "sounds", str(self)),
                    cache_dir_delete,
                )
        self.audio_service = audio_service
        self.current_tracker = None
        self.create_subcaption = create_subcaption

    def add_voiceover_text(
        self,
        text: str,
        subcaption: Optional[str] = None,
        max_subcaption_len: int = 70,
        subcaption_buff: float = 0.1,
        gain_to_background: float | None = None,
        **kwargs,
    ) -> SoundTracker:
        """Adds sound to the scene.

        Args:
            text (str): The text to be spoken.
            subcaption (Optional[str], optional): Alternative subcaption text. If not specified, `text` is chosen as the subcaption. Defaults to None.
            max_subcaption_len (int, optional): Maximum number of characters for a subcaption. Subcaptions that are longer are split into chunks that are smaller than `max_subcaption_len`. Defaults to 70.
            subcaption_buff (float, optional): The duration between split subcaption chunks in seconds. Defaults to 0.1.

        Returns:
            SoundTracker: The tracker object for the sound.
        """
        if not hasattr(self, "audio_service"):
            raise Exception("You need to call init_sound() before adding a sound.")

        dict_ = self.audio_service._wrap_generate_from_text(
            text=text, voice_id=self.voice_id, **kwargs
        )
        tracker = SoundTracker(self, dict_, self.audio_service.cache_dir, self.voice_id)

        audio_file_path = str(Path(self.audio_service.cache_dir) / dict_["final_audio"])

        if os.path.exists(audio_file_path):
            self.renderer.file_writer.add_sound(
                str(Path(self.audio_service.cache_dir) / dict_["final_audio"]),
                self.renderer.time + 0,
                None,
            )
            self.current_tracker = tracker

            if self.create_subcaption:
                if subcaption is None:
                    # Remove placeholders
                    subcaption = re.sub(r"<[^<>]+/>", "", text)

                self.add_wrapped_subcaption(
                    subcaption,
                    tracker.duration,
                    subcaption_buff=subcaption_buff,
                    max_subcaption_len=max_subcaption_len,
                )

        return tracker

    def add_wrapped_subcaption(
        self,
        subcaption: str,
        duration: float,
        subcaption_buff: float = 0.1,
        max_subcaption_len: int = 70,
    ) -> None:
        """Adds a subcaption to the scene. If the subcaption is longer than `max_subcaption_len`, it is split into chunks that are smaller than `max_subcaption_len`.

        Args:
            subcaption (str): The subcaption text.
            duration (float): The duration of the subcaption in seconds.
            max_subcaption_len (int, optional): Maximum number of characters for a subcaption. Subcaptions that are longer are split into chunks that are smaller than `max_subcaption_len`. Defaults to 70.
            subcaption_buff (float, optional): The duration between split subcaption chunks in seconds. Defaults to 0.1.
        """
        subcaption = " ".join(subcaption.split())
        n_chunk = ceil(len(subcaption) / max_subcaption_len)
        tokens = subcaption.split(" ")
        chunk_len = ceil(len(tokens) / n_chunk)
        chunks_ = list(chunks(tokens, chunk_len))
        try:
            assert len(chunks_) == n_chunk or len(chunks_) == n_chunk - 1
        except AssertionError:
            import ipdb

            ipdb.set_trace()

        subcaptions = [" ".join(i) for i in chunks_]
        subcaption_weights = [
            len(subcaption) / len("".join(subcaptions)) for subcaption in subcaptions
        ]

        current_offset = 0
        for idx, subcaption in enumerate(subcaptions):
            chunk_duration = duration * subcaption_weights[idx]
            self.add_subcaption(
                subcaption,
                duration=max(chunk_duration - subcaption_buff, 0),
                offset=current_offset,
            )
            current_offset += chunk_duration

    def say_to_wait(
        self,
        text: str | None = None,
        mobject: Mobject | None = None,
        *args,
        **kwargs,
    ) -> None:
        if text is None and len(args) == 0:
            text = "Say {}".format(self.voice_id)

        with self.voiceover(text=text, mobject=mobject) as tracker:
            if len(args):
                self.play(*args, run_time=tracker.duration)
            else:
                self.safe_wait(tracker.duration)

    def say_to_play(
        self,
        *mobject_args,
        text: str | None = None,
        mobject: Mobject | None = None,
        **kwargs,
    ):

        with self.voiceover(text=text, mobject=mobject) as tracker:
            self.play(*mobject_args, run_time=tracker.duration)

    def wait_for_voiceover(self) -> None:
        """Waits for the sound to finish."""
        if not hasattr(self, "current_tracker"):
            return
        if self.current_tracker is None:
            return

        self.safe_wait(self.current_tracker.get_remaining_duration())

    def safe_wait(self, duration: float) -> None:
        """Waits for a given duration. If the duration is less than one frame, it waits for one frame.

        Args:
            duration (float): The duration to wait for in seconds.
        """
        if duration > 1 / config["frame_rate"]:
            self.wait(duration)

    @contextmanager
    def voiceover(
        self, text: str | None = None, mobject: Mobject | None = None, **kwargs
    ) -> Generator[SoundTracker, None, None]:
        """The main function to be used for adding sound to a scene.

        Args:
            text (str, optional): The text to be spoken. Defaults to None.
            mobject (str, optional) : The Mobject to be spoken. Default to None
        Yields:
            Generator[SoundTracker, None, None]: The sound tracker object.
        """
        if text is None and mobject is None:
            raise ValueError(
                "Please specify either a sound text string and mobject path."
            )
        else:
            text, mobject = text_and_mobject(text, mobject)

        try:
            # Increment voice_id after adding a new sound
            self.voice_id += 1
            yield self.add_voiceover_text(text, mobject=mobject, **kwargs)
        finally:
            self.wait_for_voiceover()

    @override
    def render(self, preview: bool = False):
        """
        Renders this Scene.

        Parameters
        ---------
        preview
            If true, opens scene in a file viewer.
        """
        self.setup()
        try:
            self.construct()
        except EndSceneEarlyException:
            pass
        except RerunSceneException as e:
            self.remove(*self.mobjects)
            self.renderer.clear_screen()
            self.renderer.num_plays = 0
            return True
        self.tear_down()
        # We have to reset these settings in case of multiple renders.
        self.renderer.scene_finished(self)

        # Show info only if animations are rendered or to get image
        if (
            self.renderer.num_plays
            or config["format"] == "png"
            or config["save_last_frame"]
        ):
            logger.info(
                f"Rendered {str(self)}\nPlayed {self.renderer.num_plays} animations",
            )

        # If preview open up the render after rendering.
        if preview:
            config["preview"] = True

        if config["preview"] or config["show_in_file_browser"]:
            open_media_file(self.renderer.file_writer)

        if hasattr(self, "audio_service"):
            if hasattr(self.audio_service, "app_exec"):
                self.audio_service.app_exec()



"""
tracker for manim-recorder
"""

from pathlib import Path
from manim import Scene
from manim_recorder.multimedia import get_duration


AUDIO_OFFSET_RESOLUTION = 10_000_000


class SoundTracker:
    """Class to track the progress of a sound in a scene."""

    def __init__(self, scene: Scene, data: dict, cache_dir: str, voice_id: int):
        """Initializes a SoundTracker object.

        Args:
            scene (Scene): The scene to which the sound belongs.
            path (str): The path to the JSON file containing the sound data.
        """
        self.scene = scene
        self.data = data
        self.cache_dir = cache_dir
        self.voice_id = voice_id
        audio_file_path = Path(cache_dir) / self.data["final_audio"]
        self.duration = 1
        if audio_file_path.exists():
            self.duration = get_duration(Path(cache_dir) / self.data["final_audio"])
        last_t = scene.renderer.time
        if last_t is None:
            last_t = 0
        self.start_t = last_t
        self.end_t = last_t + self.duration

    def get_remaining_duration(self, buff: float = 0.0) -> float:
        """Returns the remaining duration of the sound.

        Args:
            buff (float, optional): A buffer to add to the remaining duration. Defaults to 0.

        Returns:
            int: The remaining duration of the sound in seconds.
        """
        # result= max(self.end_t - self.scene.last_t, 0)
        result = max(self.end_t - self.scene.renderer.time + buff, 0)
        return result



"""
GUI Recorder
"""

import sys
import os
import datetime
import logging
from PySide6.QtGui import QIcon, QPalette, QColor, QShortcut, QKeySequence
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QLabel,
    QVBoxLayout,
    QPushButton,
    QLineEdit,
    QWidget,
    QMessageBox,
    QComboBox,
    QStyle,
    QHBoxLayout,
    QStatusBar,
    QSizePolicy,
    QProgressBar,
    QPlainTextEdit,
    QTextEdit,
)
import numpy as np
from PySide6.QtCore import Qt, QThread, Signal, QTimer, QSize
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtSvg import QSvgRenderer
import pyqtgraph as pg
from manim_recorder.multimedia import PyAudio_
from manim_recorder.recorder.gui.util import Run_Audacity
from manim_recorder.recorder.gui.config import (
    create_label,
    Create_Button,
    WindowCenter,
    SVG_Viewer,
    SVG_Icon,
    show_message,
    setup_shortcuts,
)
from manim_recorder.helper import get_audio_basename


class Recorder(QMainWindow):
    """
    A class to represent an audio recorder application using PySide6.

    Attributes:
        recorder_service: An instance of the audio recording service.
        msg: A message to display in the UI.
        File_Saved: A boolean indicating if the audio file has been saved.
    """

    def __init__(
        self,
        msg: str = "Start",
        recorder_service=PyAudio_(),
        parent=None,
        communicator=None,
        **kwargs,
    ):
        """
        Initializes the Recorder application.

        Args:
            msg (str): The initial message to display.
            recorder_service: An instance of the audio recording service.
            parent: The parent widget.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(parent)
        self.initUI()
        self.recorder_service = recorder_service
        self.msg = msg
        self.File_Saved = False
        self.populate_device_list()
        self.waveform_chart_timer = pg.QtCore.QTimer()
        self.waveform_chart_timer.timeout.connect(self.update_plot)
        self.progress_timer = QTimer(self)
        self.progress_timer.timeout.connect(self.update_progress_bar)
        self.communicator = communicator
        if self.communicator:
            self.communicator.recorder_data.connect(self._recorder)

    def initUI(self):
        """Initializes the user interface components."""
        self.setWindowTitle("Audio Recorder")
        self.setGeometry(100, 100, 600, 400)
        WindowCenter(self, onTop=True)

        layout = QVBoxLayout()

        # Device Index List
        self.device_combo = QComboBox()
        self.device_combo.currentIndexChanged.connect(self.set_device_index)
        layout.addWidget(self.device_combo)

        # Speech Script Section
        message_layout = QVBoxLayout()

        self.message_object = SVG_Viewer(svg_file=SVG_Icon["MIC"])

        message_layout.addWidget(self.message_object, stretch=1)

        # Message Display
        self.message_label = QTextEdit(self)
        # self.message_label.setText("Message")
        self.message_label.setText("Message")
        self.setStyleSheet("color:orange; font-size: 15px;")
        # self.message_label = create_label(
        #     "Message", "color:orange; font-size: 15px;", align="c", wordwrap=True
        # )
        # self.message_label.set
        # message_layout.addStretch(1)

        message_layout.addWidget(self.message_label, stretch=1)

        layout.addLayout(message_layout)

        # Waveform Chart
        self.plot_widget = pg.PlotWidget(background="white")
        self.plot_widget.setYRange(-1, 1)
        layout.addWidget(self.plot_widget)
        self.plot_widget.setBackground(
            QColor(255, 255, 255, 0)  # White with 0% opacity
        )
        self.plot_widget.getAxis("left").setVisible(False)
        self.plot_widget.getAxis("bottom").setVisible(False)
        self.plot_widget.setMenuEnabled(False)
        self.plot_widget.setMouseTracking(False)

        # Recording Timer Display
        self.recording_timer_label = create_label(
            "00:00:00", "font-size: 30px;", align="c"
        )
        layout.addWidget(self.recording_timer_label)

        # Media Control Buttons
        media_layout = QHBoxLayout()
        self.pause_button = Create_Button(
            icon=self.style().standardIcon(QStyle.SP_MediaPause),
            func=self.__pause,
            stylesheet="modern",
            disable=True,
            toolTip="Pause (<b>t</b>)",
        )
        self.play_button = Create_Button(
            icon=self.style().standardIcon(QStyle.SP_MediaPlay),
            func=self.__play,
            stylesheet="modern",
            disable=True,
            toolTip="Play (<b>p</b>)",
        )
        self.stop_button = Create_Button(
            icon=self.style().standardIcon(QStyle.SP_MediaStop),
            func=self.__stop,
            stylesheet="modern",
            disable=True,
            toolTip="Stop (<b>s</b>)",
        )
        self.rec_button = Create_Button(
            icon=QIcon(SVG_Icon["MIC"]),
            stylesheet="modern",
            func=self.__rec,
            toolTip="Recording (<b>r</b>)",
        )

        self.save_button = Create_Button(
            icon=self.style().standardIcon(QStyle.SP_DialogSaveButton),
            stylesheet="modern",
            func=self.save_audio,
            disable=True,
            toolTip="Save (<b>Ctrl + S</b>)",
        )

        media_layout.addWidget(self.pause_button)
        media_layout.addWidget(self.play_button)
        media_layout.addWidget(self.stop_button)
        media_layout.addWidget(self.rec_button)
        media_layout.addWidget(self.save_button)

        layout.addLayout(media_layout)

        # Progress Bar
        self.progress_bar = QProgressBar(self)
        self.progress_bar.setAlignment(Qt.AlignCenter)
        self.progress_bar.setRange(0, 100)  # Set range from 0 to 100
        self.progress_bar.setTextVisible(False)
        self.progress_bar.setStyleSheet(
            """
            QProgressBar {
                border: 2px solid #555;
                border-radius: 5px;
                height: 0.5em;
                background-color: #e0e0e0;
            }
            QProgressBar::chunk {
                background-color: #76c7c0;
                border-radius: 5px;
            }
        """
        )
        layout.addWidget(self.progress_bar)

        # Additional Features Layout
        feature_layout = QHBoxLayout()

        # Run Audacity Button
        self.audacity_button = Create_Button(
            "Run Audacity",
            lambda: Run_Audacity(self.recorder_service.__str__()),
            stylesheet="modern",
            disable=True,
            toolTip="Audacity Run",
        )
        feature_layout.addWidget(self.audacity_button)

        # Accept Button
        self.accept_button = Create_Button(
            "Next",
            stylesheet="modern",
            func=self._next,
            disable=True,
            toolTip="Accept and Next (<b>a</b>)",
        )

        feature_layout.addWidget(self.accept_button)
        layout.addLayout(feature_layout)

        # Status Bar
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)
        # self.status_bar.setStyleSheet("color:red;")
        self.status_bar.showMessage("Loading ...")

        container = QWidget()
        container.setLayout(layout)
        self.setCentralWidget(container)
        setup_shortcuts(
            self,
            {
                "r": self.__rec,
                "s": self.__stop,
                "p": self.__play,
                "t": self.__pause,
                "Ctrl+S": self.save_audio,
                "a": self._next,
                "Ctrl+Q": self.close,
            },
        )
        self.curve = self.plot_widget.plot(pen="g")

    def populate_device_list(self):
        """Populates the device combo box with available audio devices."""
        self.device_combo.addItems(self.recorder_service.get_devices_name())

    def set_device_index(self, index):
        """
        Sets the currently selected device index and updates the status bar.

        Args:
            index: The index of the selected device.
        """
        self.status_bar.showMessage("Device: {}".format(index))
        self.device_index = self.device_combo.itemData(index)

    def start_timer(self):
        """Starts the timer for updating the waveform chart."""
        self.waveform_chart_timer.start(50)  # Update plot every 50 ms

    def stop_timer(self):
        """Stops the timer for updating the waveform chart."""
        self.waveform_chart_timer.stop()

    def reset_timer(self):
        """Resets the recording timer display."""
        self.stop_timer()  # Stop the timer if it's running
        self.recording_timer_label.reset()  # Reset display

    def update_plot(self):
        """Updates the waveform plot with the latest audio data."""
        if self.recorder_service.frames:
            # Convert the last chunk of frames to numpy array
            data = np.frombuffer(self.recorder_service.frames[-1], dtype=np.int16)
            data = data / 32768.0  # Normalize to [-1, 1]

            # time = np.linspace(
            #     self.recorder_service.get_recording_duration() - 1,
            #     self.recorder_service.get_recording_duration(),
            #     num=len(data),
            # )

            self.curve.setData(data)

            self.recording_timer_label.setText(
                self.recorder_service.get_recording_format_duration()
            )

    def update_progress_bar(self):
        """Updates the progress bar based on the current playback status."""
        if self.recorder_service.is_playing:
            # current_duration = (
            #     self.recorder_service.get_recording_duration()
            # )  # Total duration in seconds
            total_frames = len(self.recorder_service.frames)  # Total frames recorded
            if total_frames > 0:
                current_frame = (
                    self.recorder_service.current_playback_frame_index
                )  # This should be updated during playback
                progress = (
                    current_frame / total_frames
                ) * 100  # Calculate progress percentage
                self.progress_bar.setValue(progress)  # Update the progress bar

                if current_frame >= total_frames:
                    self.__stop()  # Automatically stop playback if the end is reached
                    self.progress_bar.setValue(100)  # Set progress bar to 100%
        else:
            self.progress_timer.stop()  # Stop the timer if not playing
            self.__stop()

    def __pause(self):
        """Pauses the audio recording if currently recording."""
        if self.recorder_service.is_recording:
            if not self.recorder_service.is_paused:
                self.recorder_service.pause_recording()
            else:
                self.recorder_service.resume_recording()
        elif self.recorder_service.is_playing:
            if not self.recorder_service.playback_paused:
                self.recorder_service.pause_playback()
            else:
                self.recorder_service.resume_playback()

    def __play(self):
        """Starts playback of the recorded audio."""
        if (
            not self.recorder_service.is_playing
            and not self.recorder_service.is_recording
        ):
            self.recorder_service.play_playback()
            self.play_button.setDisabled(True)
            self.stop_button.setDisabled(False)
            self.pause_button.setDisabled(False)
            self.rec_button.setDisabled(True)
            # Start the progress bar timer
            self.progress_timer.start(100)  # Update every 100 ms
            self.progress_bar.setValue(0)

    def __stop(self):
        """Stops the audio recording or playback."""
        if self.recorder_service.is_recording:
            self.status_bar.showMessage("Status: Re-recording")
            self.recorder_service.stop_recording()
            self.stop_timer()
        elif self.recorder_service.is_playing:
            self.recorder_service.stop_playback()
        self.stop_button.setDisabled(True)
        self.pause_button.setDisabled(True)
        self.play_button.setDisabled(False)
        self.rec_button.setDisabled(False)
        self.save_button.setDisabled(False)
        self.accept_button.setDisabled(False)

    def __rec(self):
        """Starts audio recording."""
        if not self.recorder_service.is_recording:
            self.status_bar.showMessage("Status : Recording ...")
            self.rec_button.setDisabled(True)
            self.play_button.setDisabled(True)
            self.save_button.setDisabled(True)
            self.pause_button.setDisabled(False)
            self.stop_button.setDisabled(False)

            self.recorder_service.set_device_index(self.device_index)
            self.recorder_service.set_channels(self.device_index)
            self.recorder_service.start_recording()
            self.recording_timer_label.reset()
            self.start_timer()

    def _next(self):
        if not self.File_Saved:
            if (
                show_message("Confirmation", "Do you want to save the audio file?")
                == QMessageBox.Yes
            ):
                self.save_audio()
        self.pause_button.setDisabled(True)
        self.play_button.setDisabled(True)
        self.stop_button.setDisabled(True)
        self.rec_button.setDisabled(True)
        self.save_button.setDisabled(True)
        self.accept_button.setDisabled(True)
        self.audacity_button.setDisabled(True)
        logging.info("Audio File : {}".format(self.recorder_service.__str__()))
        self.status_bar.showMessage("Loading ...")
        if self.communicator:
            self.communicator.accept.emit(self.recorder_service.__str__())

    def save_audio(self):
        """Saves the recorded audio to a file."""
        if self.recorder_service.is_recording:
            self.recorder_service.stop_recording()
        if self.recorder_service.is_playing:
            self.recorder_service.stop_playback()
        if self.recorder_service.frames:
            self.recorder_service.save_recording()
            self.File_Saved = True
            self.audacity_button.setDisabled(False)
        else:
            show_message("No Recording", "Please record before saving the file.")

    def resetUI(self):
        """Resets the user interface to its initial state."""
        self.File_Saved = False
        # self.message_label.reset()  # Clear message label
        self.message_label.setText("Message")
        self.recording_timer_label.reset()  # Reset timer display
        self.progress_bar.setValue(0)  # Reset progress bar
        self.status_bar.clearMessage()  # Clear status bar message
        self.recorder_service.frames = []
        # Enable/disable buttons to their original state
        self.pause_button.setDisabled(True)
        self.play_button.setDisabled(False)
        self.stop_button.setDisabled(True)
        self.rec_button.setDisabled(False)
        self.save_button.setDisabled(True)
        self.accept_button.setDisabled(True)
        self.audacity_button.setDisabled(True)
        self.message_object.reset()
        # Clear the waveform plot
        self.curve.setData([])  # Clear the plot data

    def _recorder(
        self,
        path: str = None,
        msg: str = None,
        voice_id: int = None,
        mobject: str = None,
    ):

        self.resetUI()
        if path is not None:
            self.recorder_service.set_filepath(path)

        if mobject is not None:
            if os.path.exists(mobject):
                self.message_object.load_svg(mobject)
        if msg is not None:
            self.message_label.setText(msg)
        if voice_id is not None:
            self.setWindowTitle("Sound ID : {}".format(voice_id))
        # if show:
        #     self.show()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    audiorecoder = PyAudio_()
    recorder = Recorder(audiorecoder)
    # recorder._recorder(path=get_audio_basename() + ".wav", msg="Start Recording")
    recorder.show()
    sys.exit(app.exec())



from pathlib import Path
from manim_recorder.helper import msg_box, get_audio_basename
from manim_recorder.recorder.base import SpeechService
from manim import logger
from manim_recorder.recorder.termux.cli import Recorder


class RecorderService(SpeechService):
    """Speech service that records from a microphone during rendering."""

    def __init__(
        self,
        encoder: str = "acc",
        channel_count: int = 1,
        sampling_rate: int = 44100,
        trim_silence_threshold: float = -40.0,
        trim_buffer_start: int = 200,
        trim_buffer_end: int = 200,
        callback_delay: float = 0.05,
        **kwargs,
    ):
        """Initialize the speech service.

        Args:
            format (int, optional): Format of the audio. Defaults to pyaudio.paInt16.
            channels (int, optional): Number of channels. Defaults to 1.
            rate (int, optional): Sampling rate. Defaults to 44100.
            chunk (int, optional): Chunk size. Defaults to 512.
            device_index (int, optional): Device index, if you don't want to choose it every time you render. Defaults to None.
            trim_silence_threshold (float, optional): Threshold for trimming silence in decibels. Defaults to -40.0 dB.
            trim_buffer_start (int, optional): Buffer duration for trimming silence at the start. Defaults to 200 ms.
            trim_buffer_end (int, optional): Buffer duration for trimming silence at the end. Defaults to 200 ms.
        """

        self.recorder = Recorder(
            encoder=encoder,
            channel_count=channel_count,
            sampling_rate=sampling_rate,
            trim_silence_threshold=trim_silence_threshold,
            trim_buffer_start=trim_buffer_start,
            trim_buffer_end=trim_buffer_end
        )

        SpeechService.__init__(self, **kwargs)

    def generate_from_text(
        self, text: str, cache_dir: str = None, path: str = None, **kwargs
    ) -> dict:
        """"""
        
        if cache_dir is None:
            cache_dir = self.cache_dir

        input_data = {
            # Remove bookmarks so that we don't record a voiceover every time we change a bookmark
            "input_text": text,
            "config": {
                "encoder": self.recorder.encoder,
                "channel_count": self.recorder.channel_count,
                "sampling_rate": self.recorder.sampling_rate,
            },
        }
        cached_result = self.get_cached_result(input_data, cache_dir)
        if cached_result is not None:
            return cached_result

        audio_path = get_audio_basename() + ".m4a" if path is None else path
        box = msg_box("Voiceover:\n\n" + text)
        self.recorder.record(str(Path(cache_dir) / audio_path), box)
        json_dict = {
            "input_data": input_data,
            "original_audio": audio_path,
        }

        return json_dict
